/*
 * foo_tta.cpp
 *
 * Description:	 TTAv1 input plug-in for Foobar 2000
 * Developed by: Alexander Djourik <sasha@iszf.irk.ru>
 *               Pavel Zhilin <pzh@iszf.irk.ru>
 *
 * Copyright (c) 2004 Alexander Djourik. All rights reserved.
 *
 */

/*
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Please see the file COPYING in this directory for full copyright
 * information.
 */

#include "foo_input_tta.h"
#include "CRC32.h"

#ifdef _DEBUG
#include <tchar.h>
#include <stdarg.h>
#define HPRINTF_MES_LENGTH 256
inline void hprintf( const _TCHAR * format, ... )
{
    _TCHAR buff[HPRINTF_MES_LENGTH];
    va_list vl;
    va_start(vl, format);
    _vsntprintf(buff, HPRINTF_MES_LENGTH, format, vl);
    va_end(vl);
    OutputDebugString(buff);
}
#else
#define hprintf __noop
#endif

#define TAGTYPES_TOTAL 5
#define DEFAULT_TAG	3

// {8CDBB31F-4B6C-4147-B452-AFE8AA349F6B}
static const GUID guid_cfg_tag_type = 
{ 0x8cdbb31f, 0x4b6c, 0x4147, { 0xb4, 0x52, 0xaf, 0xe8, 0xaa, 0x34, 0x9f, 0x6b } };
static cfg_int cfg_tagging(guid_cfg_tag_type, DEFAULT_TAG);

// {88743B34-31E6-4775-AA46-D98CAA79FB9C}
static const GUID guid_cfg_bitrate = 
{ 0x88743b34, 0x31e6, 0x4775, { 0xaa, 0x46, 0xd9, 0x8c, 0xaa, 0x79, 0xfb, 0x9c } };
static cfg_int cfg_bitrate(guid_cfg_bitrate, 1);

static const char * tag_name[TAGTYPES_TOTAL] = {
	"APEv2 tags only",
	"APEv2 and ID3v1",
	"ID3v1 tags only",
	"ID3v2 tags only",
	"ID3v2 and ID3v1"
};

// {9861B4FD-666D-4773-90FE-0C98A2CDC469}
static const GUID input_tta_preferences_guid = 
{ 0x9861b4fd, 0x666d, 0x4773, { 0x90, 0xfe, 0xc, 0x98, 0xa2, 0xcd, 0xc4, 0x69 } };

class input_tta_config : public preferences_page {
	static BOOL CALLBACK ConfigProc(HWND dialog,UINT msg,WPARAM wp,LPARAM lp) {
		HWND dlg_tag = uGetDlgItem(dialog, IDC_TAG);
		int indx;

		switch (msg) {
		case WM_INITDIALOG:
			for (indx = 0; indx < TAGTYPES_TOTAL; indx++)
				uSendMessageText(dlg_tag, CB_ADDSTRING, 0, tag_name[indx]);

			if (cfg_tagging >= 0 && cfg_tagging < TAGTYPES_TOTAL)
				uSendMessage(dlg_tag, CB_SETCURSEL, cfg_tagging, 0);
			else uSendMessage(dlg_tag, CB_SETCURSEL, 3, 0);

			CheckDlgButton(dialog, IDC_RT_BITRATE, cfg_bitrate);
			uSetDlgItemText(dialog, IDC_VERSION,
				"True Audio decoder version " TTA_VERSION "\n"
				"Foobar plug-in version " PLUGIN_VERSION);

            return TRUE;

        case WM_COMMAND:
            switch (wp) {
            case IDC_TAG | (CBN_SELCHANGE<<16):
				indx = uSendMessage(dlg_tag, CB_GETCURSEL, 0, 0);
				if (indx >= 0 && indx < TAGTYPES_TOTAL)
					cfg_tagging = indx;
                break;

            case IDC_RT_BITRATE:
                cfg_bitrate = !cfg_bitrate;
                break;
            }
            break;
        }

        return FALSE;
	}

	virtual HWND create(HWND parent) {
        return uCreateDialog(IDD_CONFIG, parent, ConfigProc);
    }
    virtual const char * get_name() { return "TTA input"; }
	virtual GUID get_guid() { return input_tta_preferences_guid; }
    virtual GUID get_parent_guid() {return guid_input;}
	virtual bool reset_query() {return false;}
	virtual void reset() {}
	virtual bool get_help_url(pfc::string_base & p_out)
	{
		return false;
	}
};

static preferences_page_factory_t<input_tta_config> g_input_tta_config_factory;

#pragma pack(1)

int input_tta::open_tta_file(tta_info *info, abort_callback & p_abort) {
	hprintf(L"open_tta_file()\n");
	tta_hdr ttahdr;
	struct {
		unsigned char	id[3];
		unsigned short	version;
		unsigned char	flags;
		unsigned char	size[4];
	} id3v2;
	__int64 datasize, origsize;
	unsigned long checksum;
	unsigned long offset = 0;

	ZeroMemory (info, sizeof(tta_info));
	infile->seek(0, p_abort);
	info->FILESIZE = infile->get_size(p_abort);

	// read ID3V2 header
	if (infile->read(&id3v2, sizeof(id3v2), p_abort) != sizeof(id3v2)) {
		info->STATE = READ_ERROR;
		return -1;
	}

	// skip ID3V2 header
	if (!memcmp (id3v2.id, "ID3", 3)) {
		if (id3v2.size[0] & 0x80) {
			info->STATE = FILE_ERROR;
			return -1;
		}
		offset = (id3v2.size[0] & 0x7f);
		offset = (offset << 7) | (id3v2.size[1] & 0x7f);
		offset = (offset << 7) | (id3v2.size[2] & 0x7f);
		offset = (offset << 7) | (id3v2.size[3] & 0x7f);
		offset += 10;
		if (id3v2.flags & (1 << 4)) offset += 10;
		infile->seek(offset, p_abort);
	} else infile->seek(0, p_abort);

	// read TTA header
	if (infile->read(&ttahdr, sizeof(tta_hdr), p_abort) != sizeof(tta_hdr)) {
		info->STATE = READ_ERROR;
		return -1;
	}

	// check for TTA3 signature
	if (ttahdr.TTAid != TTA1_SIGN) {
		info->STATE = FORMAT_ERROR;
		return -1;
	}

	checksum = crc32((unsigned char *) &ttahdr,
	sizeof(tta_hdr) - sizeof(long));
	if (checksum != ttahdr.CRC32) {
		info->STATE = FILE_ERROR;
		return -1;
	}

	// check for player supported formats
	if ((ttahdr.AudioFormat != WAVE_FORMAT_PCM &&
		ttahdr.AudioFormat != WAVE_FORMAT_IEEE_FLOAT) ||
		ttahdr.BitsPerSample > MAX_BPS ||
		ttahdr.NumChannels > MAX_NCH) {
		info->STATE = PLAYER_ERROR;
		return -1;
	}

	// fill the File Info
	info->NCH = ttahdr.NumChannels;
	info->BPS = ttahdr.BitsPerSample;
	info->BSIZE = (ttahdr.BitsPerSample + 7)/8;
	info->FORMAT = ttahdr.AudioFormat;
	info->SAMPLERATE = ttahdr.SampleRate;
	info->DATALENGTH = ttahdr.DataLength;
	info->FRAMELEN = (long) (FRAME_TIME * ttahdr.SampleRate);
	info->LENGTH = (double)ttahdr.DataLength / ttahdr.SampleRate;
	info->DATAPOS = offset;

	datasize = info->FILESIZE - info->DATAPOS;
	origsize = info->DATALENGTH * info->BSIZE * info->NCH;

	info->COMPRESS = (float) datasize / origsize;
	info->BITRATE = (long) (info->COMPRESS *
		(info->SAMPLERATE * info->NCH * info->BPS) / 1000);

	return 0;
}

#pragma pack()

void input_tta::open(service_ptr_t<file> p_filehint, const char * p_path, t_input_open_reason p_reason, abort_callback & p_abort) {	
	hprintf(L"open()\n");
	open_flags = p_reason;
	infile = p_filehint;
	input_open_file_helper(infile, p_path, p_reason, p_abort);
	seekable = infile->can_seek();
}

void input_tta::get_info(file_info & p_info,abort_callback & p_abort) {
	hprintf(L"get_info()\n");
	tta_info tmpinfo;
	char message[64];

	// we must to use a local info structure here
	int infile_pos = infile->get_position(p_abort);
	try {
		tag_processor::read_id3v2_trailing(infile, p_info, p_abort);
	} catch (exception_tag_not_found e) {
	}
	if (open_tta_file(&tmpinfo, p_abort) < 0) throw exception_io_no_data();
	infile->seek(infile_pos, p_abort);
	
	// fill decoder info
	p_info.set_length(tmpinfo.LENGTH);
	p_info.info_set_int("samplerate", tmpinfo.SAMPLERATE);
	p_info.info_set_int("channels", tmpinfo.NCH);
	p_info.info_set_int("bitspersample", tmpinfo.BPS);

	sprintf(message, "%d", tmpinfo.BITRATE);
	p_info.info_set("bitrate", message);
	sprintf(message, "%.2f", tmpinfo.COMPRESS);
	p_info.info_set("compression", message);

	p_info.info_set("codec", "True Audio");
	p_info.info_set("encoding", "lossless");
	p_info.info_set("homepage", PROJECT_URL);
}

t_filestats input_tta::get_file_stats(abort_callback & p_abort) {
	return infile->get_stats(p_abort);
}

void input_tta::decode_initialize(unsigned p_flags,abort_callback & p_abort) {
	hprintf(L"decode_initialize()\n");
	if (open_tta_file (&info, p_abort) < 0) {
		if (info.STATE != FORMAT_ERROR)
		tta_error(info.STATE);
		throw exception_io_unsupported_format();
	}

	if (player_init(p_abort) < 0) {
		tta_error(info.STATE);
		throw exception_io_unsupported_format();
	}
}

bool input_tta::decode_run(audio_chunk & p_chunk,abort_callback & p_abort) {
	//hprintf(L"decode_run(): infile=%d\n", infile->get_position(p_abort));
	while (seek_skip >= BUFFER_LENGTH) {
		int len = get_samples (pcm_buffer, BUFFER_LENGTH, p_abort);
		if (len <= 0) return false;
		seek_skip -= len;
	}

	int len	= get_samples(pcm_buffer, BUFFER_LENGTH, p_abort);
	if (len <= 0) return false;

	BYTE *data = pcm_buffer + seek_skip * info.NCH * info.BSIZE;
	len -= seek_skip, seek_skip = 0;

	if (data_float) p_chunk.set_data_32((const float *)data,
		len, info.NCH, info.SAMPLERATE);
    else p_chunk.set_data_fixedpoint(data,
		len * info.BSIZE * info.NCH, info.SAMPLERATE, info.NCH, info.BPS, audio_chunk::g_guess_channel_config(info.NCH));

	return true;
}

void input_tta::decode_seek(double p_seconds,abort_callback & p_abort) {
	hprintf(L"decode_seek()\n");
	if (!seekable) throw exception_io_object_not_seekable();
	unsigned long pos = (long)(p_seconds / FRAME_TIME);

	if (set_position (pos, p_abort) != 0) throw exception_io_object_not_seekable();
	seek_skip = (unsigned long)((p_seconds - (pos * FRAME_TIME)) * info.SAMPLERATE + 0.5);
}

bool input_tta::decode_can_seek() {
	return infile->can_seek();
}

bool input_tta::decode_get_dynamic_info(file_info & p_out, double & p_timestamp_delta) {
	if (!cfg_bitrate) return false;
	if (bitrate == p_out.info_get_bitrate_vbr())
		return false;
	p_out.info_set_bitrate_vbr(bitrate);
	return true;
}

bool input_tta::decode_get_dynamic_info_track(file_info & p_out, double & p_timestamp_delta) {
	return decode_get_dynamic_info(p_out, p_timestamp_delta);
}

void input_tta::decode_on_idle(abort_callback & p_abort) {
	infile->on_idle(p_abort);
}

void input_tta::retag(const file_info & p_info,abort_callback & p_abort) {
	hprintf(L"retag()\n");
	if (!seekable) throw exception_io_unsupported_format();

	__int64 len = infile->get_size(p_abort);
	if (len <= 0) throw exception_io_unsupported_format();
	
	switch(cfg_tagging) {
	case 0:
		tag_processor::write_apev2(infile, p_info, p_abort);
		break;
	case 1:
		tag_processor::write_apev2_id3v1(infile, p_info, p_abort);
		break;
	case 2:
		tag_processor::write_id3v1(infile, p_info, p_abort);
		break;
	case 3:
		tag_processor::write_id3v2(infile, p_info, p_abort);
		break;
	case 4:
		tag_processor::write_id3v2_id3v1(infile, p_info, p_abort);
		break;
	case 5:
		tag_processor::write_id3v2_apev2(infile, p_info, p_abort);
		break;
	}
}

bool input_tta::g_is_our_content_type(const char * p_content_type) {
	return false;
}

bool input_tta::g_is_our_path(const char * p_path,const char * p_extension) {
	return (stricmp_utf8(p_extension, "TTA") == 0);
}

void input_tta::tta_error (int error) {
	pfc::string8 message;

	switch (error) {
	case OPEN_ERROR:	message = "Can't open file"; break;
	case FORMAT_ERROR:	message = "Unknown TTA format version"; break;
	case PLAYER_ERROR:	message = "Not supported file format"; break;
	case FILE_ERROR:	message = "File is corrupted"; break;
	case READ_ERROR:	message = "Can't read from file"; break;
	default:			message = "Unknown TTA decoder error"; break;
	}
	
	console::error (message);
}

static input_singletrack_factory_t<input_tta> g_input_tta_factory;
//static service_factory_single_t<config, config_tta> foo2;

DECLARE_COMPONENT_VERSION("TTA Audio Decoder (unofficial)", PLUGIN_VERSION, COPYRIGHT);
DECLARE_FILE_TYPE("TTA audio file", "*.TTA");

///////////////////////////////////////////////////////////////////////
//	Description:	 TTAv1 lossless audio decoder.                   //
//	Copyright (c) 1999-2004 Alexander Djourik. All rights reserved.  //
///////////////////////////////////////////////////////////////////////

///////////////////// constants and definitions ///////////////////////

#define PREDICTOR1(x, k)	((long)((((__int64)x << k) - x) >> k))
#define DEC(x)			(((x)&1)?(++(x)>>1):(-(x)>>1))
#define WRITE_BUFFER(x, bsize, out) { \
	*out++ = (bsize == 1)? (*x + 0x80): *x; \
	if (bsize > 1) *out++ = (byte)(*x >> 8); \
	if (bsize > 2) *out++ = (byte)(*x >> 16); }

#define SWAP16(x) (\
(((x)&(1<< 0))?(1<<15):0) | \
(((x)&(1<< 1))?(1<<14):0) | \
(((x)&(1<< 2))?(1<<13):0) | \
(((x)&(1<< 3))?(1<<12):0) | \
(((x)&(1<< 4))?(1<<11):0) | \
(((x)&(1<< 5))?(1<<10):0) | \
(((x)&(1<< 6))?(1<< 9):0) | \
(((x)&(1<< 7))?(1<< 8):0) | \
(((x)&(1<< 8))?(1<< 7):0) | \
(((x)&(1<< 9))?(1<< 6):0) | \
(((x)&(1<<10))?(1<< 5):0) | \
(((x)&(1<<11))?(1<< 4):0) | \
(((x)&(1<<12))?(1<< 3):0) | \
(((x)&(1<<13))?(1<< 2):0) | \
(((x)&(1<<14))?(1<< 1):0) | \
(((x)&(1<<15))?(1<< 0):0))

const unsigned long bit_mask[] = {
    0x00000000, 0x00000001, 0x00000003, 0x00000007,
    0x0000000f, 0x0000001f, 0x0000003f, 0x0000007f,
    0x000000ff, 0x000001ff, 0x000003ff, 0x000007ff,
    0x00000fff, 0x00001fff, 0x00003fff, 0x00007fff,
    0x0000ffff, 0x0001ffff, 0x0003ffff, 0x0007ffff,
    0x000fffff, 0x001fffff, 0x003fffff, 0x007fffff,
    0x00ffffff, 0x01ffffff, 0x03ffffff, 0x07ffffff,
    0x0fffffff, 0x1fffffff, 0x3fffffff, 0x7fffffff,
    0xffffffff
};

const unsigned long bit_shift[] = {
    0x00000001, 0x00000002, 0x00000004, 0x00000008,
    0x00000010, 0x00000020, 0x00000040, 0x00000080,
    0x00000100, 0x00000200, 0x00000400, 0x00000800,
    0x00001000, 0x00002000, 0x00004000, 0x00008000,
    0x00010000, 0x00020000, 0x00040000, 0x00080000,
    0x00100000, 0x00200000, 0x00400000, 0x00800000,
    0x01000000, 0x02000000, 0x04000000, 0x08000000,
    0x10000000, 0x20000000, 0x40000000, 0x80000000,
    0x80000000, 0x80000000, 0x80000000, 0x80000000,
    0x80000000, 0x80000000, 0x80000000, 0x80000000
};

const unsigned long *shift_16 = bit_shift + 4;

///////////////////////// bit operations //////////////////////////////

void  input_tta::init_buffer_read() {
    frame_crc32 = 0xFFFFFFFFUL;
    bit_count = bit_cache = 0;
    bitpos = (iso_buffer + BUFFER_SIZE);
}

__inline void input_tta::get_binary(unsigned long *value, unsigned long bits, abort_callback & p_abort) {
    while (bit_count < bits) {
		if (bitpos == (iso_buffer + BUFFER_SIZE)) {
			if (!infile->read(iso_buffer, BUFFER_SIZE, p_abort)) {
				info.STATE = READ_ERROR;
				return;
			}
			bitpos = iso_buffer;
		}

		UPDATE_CRC32(*bitpos, frame_crc32);
		bit_cache |= *bitpos << bit_count;
		bit_count += 8;
		bitpos++;
    }

    *value = bit_cache & bit_mask[bits];
    bit_cache >>= bits;
    bit_count -= bits;
    bit_cache &= bit_mask[bit_count];
}

__inline void input_tta::get_unary(unsigned long *value, abort_callback & p_abort) {
    *value = 0;

    while (!(bit_cache ^ bit_mask[bit_count])) {
		if (bitpos == (iso_buffer + BUFFER_SIZE)) {
			if (!infile->read(iso_buffer, BUFFER_SIZE, p_abort)) {
				info.STATE = READ_ERROR;
				return;
			}
			bitpos = iso_buffer;
		}

		*value += bit_count;
		bit_cache = *bitpos++;
		UPDATE_CRC32(bit_cache, frame_crc32);
		bit_count = 8;
    }

    while (bit_cache & 1) {
		(*value)++;
		bit_cache >>= 1;
		bit_count--;
    }

    bit_cache >>= 1;
    bit_count--;
}

__inline int input_tta::done_buffer_read(abort_callback & p_abort) {
    unsigned long crc32, rbytes, result;
    frame_crc32 ^= 0xFFFFFFFFUL;

    rbytes = (iso_buffer + BUFFER_SIZE) - bitpos;
    if (rbytes < sizeof(long)) {
		memcpy(iso_buffer, bitpos, 4); 
		if (!infile->read(iso_buffer + rbytes,
			BUFFER_SIZE - rbytes, p_abort)) {
			info.STATE = READ_ERROR;
			return 0;
		}
		bitpos = iso_buffer;
    }

	memcpy(&crc32, bitpos, 4);
    bitpos += sizeof(long);
    result = (crc32 != frame_crc32);

    bit_cache = bit_count = 0;
    frame_crc32 = 0xFFFFFFFFUL;

    // calculate dynamic bitrate
    if (data_pos && data_pos < fframes) {
			rbytes = seek_table[data_pos] -
				seek_table[data_pos - 1];
			bitrate = (rbytes << 3)/
				(long)(1000 * FRAME_TIME);
    }

    return result;
}

//////////////////////// TTA hybrid filter ////////////////////////////

///////// Filter Settings //////////
static long flt_set [4][2] = {
	{10,1}, {9,1}, {10,1}, {12,0}
};

__inline void memshl (long *pA, long *pB) {
	*pA++ = *pB++;
	*pA++ = *pB++;
	*pA++ = *pB++;
	*pA++ = *pB++;
	*pA++ = *pB++;
	*pA++ = *pB++;
	*pA++ = *pB++;
	*pA   = *pB;
}

__inline void hybrid_filter (fltst *fs, long *in) {
	long *pA = fs->dl;
	long *pB = fs->qm;
	long *pM = fs->dx;
	long sum = fs->round;

	if (!fs->error) {
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++;
		sum += *pA++ * *pB, pB++; pM += 8;
	} else if (fs->error < 0) {
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
		sum += *pA++ * (*pB -= *pM++), pB++;
	} else {
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
		sum += *pA++ * (*pB += *pM++), pB++;
	}

	*(pM-0) = ((*(pA-1) >> 30) | 1) << 2;
	*(pM-1) = ((*(pA-2) >> 30) | 1) << 1;
	*(pM-2) = ((*(pA-3) >> 30) | 1) << 1;
	*(pM-3) = ((*(pA-4) >> 30) | 1);

	fs->error = *in;
	*in += (sum >> fs->shift);
	*pA = *in;

	if (fs->mutex) {
		*(pA-1) = *(pA-0) - *(pA-1);
		*(pA-2) = *(pA-1) - *(pA-2);
		*(pA-3) = *(pA-2) - *(pA-3);
	}

	memshl (fs->dl, fs->dl + 1);
	memshl (fs->dx, fs->dx + 1);
}

static void filter_init (fltst *fs, long shift, long mode) {
	ZeroMemory (fs, sizeof(fltst));
	fs->shift = shift;
	fs->round = 1 << (shift - 1);
	fs->mutex = mode;
}

//////////////////////// decoder functions ////////////////////////////

static void rice_init(adapt *rice, unsigned long k0, unsigned long k1) {
    rice->k0 = k0;
    rice->k1 = k1;
    rice->sum0 = shift_16[k0];
    rice->sum1 = shift_16[k1];
}

void input_tta::decoder_init(decoder *tta, long nch, long byte_size) {
    long *fset = flt_set[byte_size - 1];
    long i;

    for (i = 0; i < (nch << data_float); i++) {
		filter_init(&tta[i].fst, fset[0], fset[1]);
		rice_init(&tta[i].rice, 10, 10);
		tta[i].last = 0;
    }
}

static void seek_table_init (unsigned long *seek_table,
	unsigned long len, unsigned long data_offset) {
	unsigned long *st, frame_len;

	for (st = seek_table; st < (seek_table + len); st++) {
		frame_len = *st; *st = data_offset;
		data_offset += frame_len;
	}
}

int input_tta::set_position (unsigned long pos, abort_callback & p_abort) {
	unsigned long seek_pos;

	if (pos >= fframes) return 0;
	if (!st_state) {
		info.STATE = FILE_ERROR;
		return -1;
	}

	seek_pos = info.DATAPOS + seek_table[data_pos = pos];
	infile->seek_ex(seek_pos, file::t_seek_mode::seek_from_beginning, p_abort);

	data_cur = 0;
	framelen = 0;

	// init bit reader
	init_buffer_read();

	return 0;
}

int input_tta::player_init (abort_callback & p_abort) {
	unsigned long checksum;
	unsigned long data_offset;
	unsigned long st_size;

	framelen  = 0;
	data_pos  = 0;
	data_cur  = 0;
	bitrate   = 0;
	seek_skip = 0;

	lastlen = info.DATALENGTH % info.FRAMELEN;
	fframes = info.DATALENGTH / info.FRAMELEN + (lastlen ? 1:0);

	seek_table = new unsigned long[fframes + 1];
	st_size = (fframes + 1) * sizeof(long);

	// read seek table
	if (infile->read(seek_table, st_size, p_abort) != st_size) {
		info.STATE = READ_ERROR;
		return -1;
	}

	checksum = crc32((unsigned char *) seek_table, fframes * sizeof(long));
	st_state = (checksum == seek_table[fframes]);
	data_offset = sizeof(tta_hdr) + st_size;

	// init seek table
	seek_table_init(seek_table, fframes, data_offset);

	// init bit reader
	init_buffer_read();

	data_float = (info.FORMAT == WAVE_FORMAT_IEEE_FLOAT);
	maxvalue = (1UL << info.BPS) - 1;

	return 0;
}

int input_tta::get_samples (BYTE *buffer, long count, abort_callback & p_abort) {
	unsigned long k, depth, unary, binary;
	long buffer_size = count * info.BSIZE * info.NCH;
	BYTE *p = buffer;
	decoder *dec = tta;
	long *prev = cache;
	long value;
	int res, flag;

	for (res = flag = 0; p < buffer + buffer_size;) {
		fltst *fst = &dec->fst;
		adapt *rice = &dec->rice;
		long  *last = &dec->last;

		if (data_cur == framelen) {
			if (data_pos == fframes) break;
			if (framelen && done_buffer_read(p_abort)) {
				if (set_position(data_pos, p_abort) < 0)
					return -1;
				if (res) break;
			}

			if (data_pos == fframes - 1 && lastlen)
				framelen = lastlen;
			else framelen = info.FRAMELEN;

			decoder_init(tta, info.NCH, info.BSIZE);
			data_pos++; data_cur = 0;
		}

		// decode Rice unsigned
		get_unary(&unary, p_abort);

		switch (unary) {
		case 0: depth = 0; k = rice->k0; break;
		default:
				depth = 1; k = rice->k1;
				unary--;
		}

		if (k) {
			get_binary(&binary, k, p_abort);
			value = (unary << k) + binary;
		} else value = unary;

		switch (depth) {
		case 1: 
			rice->sum1 += value - (rice->sum1 >> 4);
			if (rice->k1 > 0 && rice->sum1 < shift_16[rice->k1])
				rice->k1--;
			else if (rice->sum1 > shift_16[rice->k1 + 1])
				rice->k1++;
			value += bit_shift[rice->k0];
		default:
			rice->sum0 += value - (rice->sum0 >> 4);
			if (rice->k0 > 0 && rice->sum0 < shift_16[rice->k0])
				rice->k0--;
			else if (rice->sum0 > shift_16[rice->k0 + 1])
			rice->k0++;
		}

		value = DEC(value);

		// decompress stage 1: adaptive hybrid filter
		hybrid_filter(fst, &value);

		// decompress stage 2: fixed order 1 prediction
		switch (info.BSIZE) {
		case 1: value += PREDICTOR1(*last, 4); break;	// bps 8
		case 2: value += PREDICTOR1(*last, 5); break;	// bps 16
		case 3: value += PREDICTOR1(*last, 5); break;	// bps 24
		case 4: value += *last; break;		// bps 32
		} *last = value;

		// check for errors
		if (!data_float && abs(value) > maxvalue) {
			unsigned long tail =
				buffer_size / (info.BSIZE * info.NCH) - res;
			ZeroMemory(buffer, buffer_size);
			data_cur += tail; res += tail;
			break;
		}

		if (data_float && flag) {
			unsigned long negative = value & 0x80000000;
			unsigned long data_hi = *prev;
			unsigned long data_lo = abs(value) - 1;

			data_hi += (data_hi || data_lo) ? 0x3F80 : 0;
			value = (data_hi << 16) | SWAP16(data_lo) | negative;

			*p++ = (BYTE) value;
			if (info.BSIZE > 1) *p++ = (BYTE)(value >> 8);
			if (info.BSIZE > 2) *p++ = (BYTE)(value >> 16);
			if (info.BSIZE > 3) *p++ = (BYTE)(value >> 24);

			flag = 0;
		} else flag = 1;

		if (dec < tta + (info.NCH << data_float) - 1) {
			if (!data_float) *prev++ = value;
			else *prev = value;
			dec++;
		} else {
			if (!data_float) {
				*prev = value;
				if (info.NCH > 1) {
					long *r = prev - 1;
					for (*prev += *r/2; r >= cache; r--)
						*r = *(r + 1) - *r;
					for (r = cache; r < prev; r++)
						WRITE_BUFFER(r, info.BSIZE, p)
				}
				WRITE_BUFFER(prev, info.BSIZE, p)
				prev = cache;
			}
			data_cur++; res++;
			dec = tta;
		}
	}

	return res;
}

/* eof */

